/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2020 Realtek Technologies Co., Ltd */

#include <sys/queue.h>
#include <stdio.h>
#include <errno.h>
#include <stdint.h>
#include <stdarg.h>

#include <rte_ether.h>
#include <ethdev_driver.h>

#include "r8125_ethdev.h"
#include "r8125_hw.h"

#define NO_BASE_ADDRESS 0x0

void rtl8125_mac_ocp_write(struct rtl_hw *hw, u16 addr, u16 value)
{
        u32 data32;

        data32 = addr/2;
        data32 <<= OCPR_Addr_Reg_shift;
        data32 += value;
        data32 |= OCPR_Write;

        RTL_W32(hw, MACOCP, data32);
}

u16 rtl8125_mac_ocp_read(struct rtl_hw *hw, u16 addr)
{
        u32 data32;
        u16 data16 = 0;

        data32 = addr/2;
        data32 <<= OCPR_Addr_Reg_shift;

        RTL_W32(hw, MACOCP, data32);
        data16 = (u16)RTL_R32(hw, MACOCP);

        return data16;
}

void ClearAndSetPCIePhyBit(struct rtl_hw *hw, u8 addr, u16 clearmask, u16 setmask)
{
        u16 EphyValue;

        EphyValue = rtl8125_ephy_read(hw, addr);
        EphyValue &= ~clearmask;
        EphyValue |= setmask;
        rtl8125_ephy_write(hw, addr, EphyValue);
}

void ClearPCIePhyBit(struct rtl_hw *hw, u8 addr, u16 mask)
{
        ClearAndSetPCIePhyBit( hw, addr, mask, 0);
}

void SetPCIePhyBit( struct rtl_hw *hw, u8 addr, u16 mask)
{
        ClearAndSetPCIePhyBit( hw, addr, 0, mask);
}

void rtl8125_ephy_write(struct rtl_hw *hw, int addr, int value)
{
        int i;

        RTL_W32(hw, EPHYAR,
                EPHYAR_Write |
                (addr & EPHYAR_Reg_Mask_v2) << EPHYAR_Reg_shift |
                (value & EPHYAR_Data_Mask));

        for (i = 0; i < 10; i++) {
                udelay(100);

                /* Check if the RTL8125 has completed EPHY write */
                if (!(RTL_R32(hw, EPHYAR) & EPHYAR_Flag))
                        break;
        }

        udelay(20);
}

u16 rtl8125_ephy_read(struct rtl_hw *hw, int addr)
{
        int i;
        u16 value = 0xffff;

        RTL_W32(hw, EPHYAR,
                EPHYAR_Read | (addr & EPHYAR_Reg_Mask_v2) << EPHYAR_Reg_shift);

        for (i = 0; i < 10; i++) {
                udelay(100);

                /* Check if the RTL8125 has completed EPHY read */
                if (RTL_R32(hw, EPHYAR) & EPHYAR_Flag) {
                        value = (u16) (RTL_R32(hw, EPHYAR) & EPHYAR_Data_Mask);
                        break;
                }
        }

        udelay(20);

        return value;
}

static u32 rtl8125_eri_read_with_oob_base_address(struct rtl_hw *hw, int addr, int len, int type, const u32 base_address)
{
        int i, val_shift, shift = 0;
        u32 value1 = 0, value2 = 0, mask;
        u32 eri_cmd;
        const u32 transformed_base_address = ((base_address & 0x00FFF000) << 6) | (base_address & 0x000FFF);

        if (len > 4 || len <= 0)
                return -1;

        while (len > 0) {
                val_shift = addr % ERIAR_Addr_Align;
                addr = addr & ~0x3;

                eri_cmd = ERIAR_Read |
                          transformed_base_address |
                          type << ERIAR_Type_shift |
                          ERIAR_ByteEn << ERIAR_ByteEn_shift |
                          (addr & 0x0FFF);
                if (addr & 0xF000) {
                        u32 tmp;

                        tmp = addr & 0xF000;
                        tmp >>= 12;
                        eri_cmd |= (tmp << 20) & 0x00F00000;
                }

                RTL_W32(hw, ERIAR, eri_cmd);

                for (i = 0; i < 10; i++) {
                        udelay(100);

                        /* Check if the RTL8125 has completed ERI read */
                        if (RTL_R32(hw, ERIAR) & ERIAR_Flag)
                                break;
                }

                if (len == 1)       mask = (0xFF << (val_shift * 8)) & 0xFFFFFFFF;
                else if (len == 2)  mask = (0xFFFF << (val_shift * 8)) & 0xFFFFFFFF;
                else if (len == 3)  mask = (0xFFFFFF << (val_shift * 8)) & 0xFFFFFFFF;
                else            mask = (0xFFFFFFFF << (val_shift * 8)) & 0xFFFFFFFF;

                value1 = RTL_R32(hw, ERIDR) & mask;
                value2 |= (value1 >> val_shift * 8) << shift * 8;

                if (len <= 4 - val_shift) {
                        len = 0;
                } else {
                        len -= (4 - val_shift);
                        shift = 4 - val_shift;
                        addr += 4;
                }
        }

        udelay(20);

        return value2;
}

u32 rtl8125_eri_read(struct rtl_hw *hw, int addr, int len, int type)
{
        return rtl8125_eri_read_with_oob_base_address(hw, addr, len, type, 0);
}

static int rtl8125_eri_write_with_oob_base_address(struct rtl_hw *hw, int addr, int len, u32 value, int type, const u32 base_address)
{
        int i, val_shift, shift = 0;
        u32 value1 = 0, mask;
        u32 eri_cmd;
        const u32 transformed_base_address = ((base_address & 0x00FFF000) << 6) | (base_address & 0x000FFF);

        if (len > 4 || len <= 0)
                return -1;

        while (len > 0) {
                val_shift = addr % ERIAR_Addr_Align;
                addr = addr & ~0x3;

                if (len == 1)       mask = (0xFF << (val_shift * 8)) & 0xFFFFFFFF;
                else if (len == 2)  mask = (0xFFFF << (val_shift * 8)) & 0xFFFFFFFF;
                else if (len == 3)  mask = (0xFFFFFF << (val_shift * 8)) & 0xFFFFFFFF;
                else            mask = (0xFFFFFFFF << (val_shift * 8)) & 0xFFFFFFFF;

                value1 = rtl8125_eri_read_with_oob_base_address(hw, addr, 4, type, base_address) & ~mask;
                value1 |= ((value << val_shift * 8) >> shift * 8);

                RTL_W32(hw, ERIDR, value1);

                eri_cmd = ERIAR_Write |
                          transformed_base_address |
                          type << ERIAR_Type_shift |
                          ERIAR_ByteEn << ERIAR_ByteEn_shift |
                          (addr & 0x0FFF);
                if (addr & 0xF000) {
                        u32 tmp;

                        tmp = addr & 0xF000;
                        tmp >>= 12;
                        eri_cmd |= (tmp << 20) & 0x00F00000;
                }

                RTL_W32(hw, ERIAR, eri_cmd);

                for (i = 0; i < 10; i++) {
                        udelay(100);

                        /* Check if the RTL8125 has completed ERI write */
                        if (!(RTL_R32(hw, ERIAR) & ERIAR_Flag))
                                break;
                }

                if (len <= 4 - val_shift) {
                        len = 0;
                } else {
                        len -= (4 - val_shift);
                        shift = 4 - val_shift;
                        addr += 4;
                }
        }

        udelay(20);

        return 0;
}

int rtl8125_eri_write(struct rtl_hw *hw, int addr, int len, u32 value, int type)
{
        return rtl8125_eri_write_with_oob_base_address(hw, addr, len, value, type, NO_BASE_ADDRESS);
}

u32 rtl8125_csi_read(struct rtl_hw *hw, u32 addr)
{
        u32 cmd;
        int i;
        u32 value = 0;

        cmd = CSIAR_Read | CSIAR_ByteEn << CSIAR_ByteEn_shift | (addr & CSIAR_Addr_Mask);

        RTL_W32(hw, CSIAR, cmd);

        for (i = 0; i < 10; i++) {
                udelay(100);

                /* Check if the RTL8125 has completed CSI read */
                if (RTL_R32(hw, CSIAR) & CSIAR_Flag) {
                        value = (u32)RTL_R32(hw, CSIDR);
                        break;
                }
        }

        udelay(20);

        return value;
}

void rtl8125_get_mac_version(struct rtl_hw *hw)
{
        u32 reg,val32;
        u32 ICVerID;

        val32 = RTL_R32(hw, TxConfig);
        reg = val32 & 0x7c800000;
        ICVerID = val32 & 0x00700000;

        switch (reg) {
        case 0x60800000:
                if (ICVerID == 0x00000000) {
                        hw->mcfg = CFG_METHOD_2;
                } else if (ICVerID == 0x100000) {
                        hw->mcfg = CFG_METHOD_3;
                } else {
                        hw->mcfg = CFG_METHOD_3;
                        hw->HwIcVerUnknown = TRUE;
                }

                hw->efuse_ver = EFUSE_SUPPORT_V4;
                break;
        case 0x64000000:
                if (ICVerID == 0x00000000) {
                        hw->mcfg = CFG_METHOD_4;
                } else if (ICVerID == 0x100000) {
                        hw->mcfg = CFG_METHOD_5;
                } else {
                        hw->mcfg = CFG_METHOD_5;
                        hw->HwIcVerUnknown = TRUE;
                }

                hw->efuse_ver = EFUSE_SUPPORT_V4;
                break;
        default:
                printf("unknown chip version (%x)\n",reg);
                hw->mcfg = CFG_METHOD_DEFAULT;
                hw->HwIcVerUnknown = TRUE;
                hw->efuse_ver = EFUSE_NOT_SUPPORT;
                break;
        }
}

void rtl8125_enable_cfg9346_write(struct rtl_hw *hw)
{
        RTL_W8(hw, Cfg9346, RTL_R8(hw, Cfg9346) | Cfg9346_Unlock);
}

void rtl8125_disable_cfg9346_write(struct rtl_hw *hw)
{
        RTL_W8(hw, Cfg9346, RTL_R8(hw, Cfg9346) & ~Cfg9346_Unlock);
}

void rtl8125_rar_set(struct rtl_hw *hw, uint8_t *addr)
{
        uint32_t rar_low = 0;
        uint32_t rar_high = 0;

        rar_low = ((uint32_t) addr[0] |
                   ((uint32_t) addr[1] << 8) |
                   ((uint32_t) addr[2] << 16) |
                   ((uint32_t) addr[3] << 24));

        rar_high = ((uint32_t) addr[4] |
                    ((uint32_t) addr[5] << 8));

        rtl8125_enable_cfg9346_write(hw);
        RTL_W32(hw, MAC0, rar_low);
        RTL_W32(hw, MAC4, rar_high);

        rtl8125_disable_cfg9346_write(hw);
}

int rtl8125_get_mac_address(struct rtl_hw *hw, struct rte_ether_addr *ea)
{
        int i;
        u8 mac_addr[MAC_ADDR_LEN];

        for (i = 0; i < MAC_ADDR_LEN; i++)
                mac_addr[i] = RTL_R8(hw, MAC0 + i);

        switch (hw->mcfg) {
        case CFG_METHOD_2 ... CFG_METHOD_5:
                *(u32*)&mac_addr[0] = RTL_R32(hw, BACKUP_ADDR0_8125);
                *(u16*)&mac_addr[4] = RTL_R16(hw, BACKUP_ADDR1_8125);
                break;
        default:
                break;
        }

        rte_ether_addr_copy((struct rte_ether_addr *)mac_addr, ea);

        return 0;

}

void rtl8125_set_hw_ops(struct rtl_hw *hw)
{
        switch (hw->mcfg) {
        //8125A
        case CFG_METHOD_2:
        case CFG_METHOD_3:
                hw->hw_ops = rtl8125a_ops;
                break;
        //8125B
        case CFG_METHOD_4:
        case CFG_METHOD_5:
                hw->hw_ops = rtl8125b_ops;
                break;
        default:
                hw->hw_ops = rtl8125b_ops;
                break;
        }
}

void rtl8125_wait_txrx_fifo_empty(struct rtl_hw *hw)
{
        int i;
        switch (hw->mcfg) {
        case CFG_METHOD_2 ... CFG_METHOD_5:
                for (i = 0; i < 3000; i++) {
                        udelay(50);
                        if ((RTL_R8(hw, MCUCmd_reg) & (Txfifo_empty | Rxfifo_empty)) == (Txfifo_empty | Rxfifo_empty))
                                break;
                }
                break;
        }

        switch (hw->mcfg) {
        case CFG_METHOD_4:
        case CFG_METHOD_5:
                for (i = 0; i < 3000; i++) {
                        udelay(50);
                        if ((RTL_R16(hw, IntrMitigate) & (BIT_0 | BIT_1 | BIT_8)) == (BIT_0 | BIT_1 | BIT_8))
                                break;
                }
                break;
        }

}

static void rtl8125_wait_ll_share_fifo_ready(struct rtl_hw *hw)
{
        int i;

        for (i = 0; i < 10; i++) {
                udelay(100);
                if (RTL_R16(hw, 0xD2) & BIT_9)
                        break;
        }
}

void rtl8125_disable_now_is_oob(struct rtl_hw *hw)
{
        switch (hw->mcfg) {
        case CFG_METHOD_2 ... CFG_METHOD_5:
                if ( hw->HwSuppNowIsOobVer == 1 )
                        RTL_W8(hw, MCUCmd_reg, RTL_R8(hw, MCUCmd_reg) & ~Now_is_oob);
                break;
        }

}

void rtl8125_init_ll_share_fifo(struct rtl_hw *hw)
{
        u16 data16;
        switch (hw->mcfg) {
        case CFG_METHOD_2 ... CFG_METHOD_5:
                data16 = rtl8125_mac_ocp_read(hw, 0xE8DE) & ~BIT_14;
                rtl8125_mac_ocp_write(hw, 0xE8DE, data16);
                rtl8125_wait_ll_share_fifo_ready(hw);

                rtl8125_mac_ocp_write(hw, 0xC0AA, 0x07D0);
                rtl8125_mac_ocp_write(hw, 0xC0A6, 0x01B5);
                rtl8125_mac_ocp_write(hw, 0xC01E, 0x5555);

                rtl8125_wait_ll_share_fifo_ready(hw);
                break;
        }
}

void
rtl8125_hw_disable_mac_mcu_bps(struct rtl_hw *hw)
{
        u16 regAddr;

        switch (hw->mcfg) {
        case CFG_METHOD_2 ... CFG_METHOD_5:
                rtl8125_enable_cfg9346_write(hw);
                RTL_W8(hw, Config5, RTL_R8(hw, Config5) & ~BIT_0);
                RTL_W8(hw, Config2, RTL_R8(hw, Config2) & ~BIT_7);
                rtl8125_disable_cfg9346_write(hw);
                break;
        }

        switch (hw->mcfg) {
        case CFG_METHOD_2 ... CFG_METHOD_5:
                rtl8125_mac_ocp_write(hw, 0xFC38, 0x0000);
                break;
        }

        switch (hw->mcfg) {
        case CFG_METHOD_2 ... CFG_METHOD_5:
                for (regAddr = 0xFC28; regAddr < 0xFC48; regAddr += 2) {
                        rtl8125_mac_ocp_write(hw, regAddr, 0x0000);
                }

                mdelay(3);

                rtl8125_mac_ocp_write(hw, 0xFC26, 0x0000);
                break;
        }
}

void
rtl8125_enable_rxdvgate(struct rtl_hw *hw)
{
        switch (hw->mcfg) {
        case CFG_METHOD_2 ... CFG_METHOD_5:
                RTL_W8(hw, 0xF2, RTL_R8(hw, 0xF2) | BIT_3);
                mdelay(2);
        }
}

void
rtl8125_disable_rxdvgate(struct rtl_hw *hw)
{
        switch (hw->mcfg) {
        case CFG_METHOD_2 ... CFG_METHOD_5:
                RTL_W8(hw, 0xF2, RTL_R8(hw, 0xF2) & ~BIT_3);
                mdelay(2);
        }
}

void
rtl8125_nic_reset(struct rtl_hw *hw)
{
        int i;

        RTL_W32(hw, RxConfig, (RX_DMA_BURST << RxCfgDMAShift));

        rtl8125_enable_rxdvgate(hw);

        rtl8125_wait_txrx_fifo_empty(hw);

        switch (hw->mcfg) {
        case CFG_METHOD_2 ... CFG_METHOD_5:
                mdelay(2);
                break;
        }

        /* Soft reset the chip. */
        RTL_W8(hw, ChipCmd, CmdReset);

        /* Check that the chip has finished the reset. */
        for (i = 100; i > 0; i--) {
                udelay(100);
                if ((RTL_R8(hw, ChipCmd) & CmdReset) == 0)
                        break;
        }
        //printf("/********rtl8125_nic_reset****/\n");
        //8111DP (CFG11) has extra action, omit it currently
}

static void
rtl8125_exit_realwow(struct rtl_hw *hw)
{
        //Disable realwow  function
        switch (hw->mcfg) {
        case CFG_METHOD_2 ... CFG_METHOD_5:
                rtl8125_mac_ocp_write(hw, 0xC0BC, 0x00FF);
                break;
        }
}

void
rtl8125_exit_oob(struct rtl_hw *hw)
{
        RTL_W32(hw, RxConfig, RTL_R32(hw, RxConfig) & ~(AcceptErr | AcceptRunt | AcceptBroadcast | AcceptMulticast | AcceptMyPhys |  AcceptAllPhys));

        //omit dash code

        rtl8125_exit_realwow(hw);

        rtl8125_nic_reset(hw);

        rtl8125_disable_now_is_oob(hw);
        rtl8125_init_ll_share_fifo(hw);

        //wait ups resume (phy state 2)
        //switch (hw->mcfg) {
        //case CFG_METHOD_29 ... CFG_METHOD_33:
        //        if (rtl8125_is_ups_resume(hw)) {
        //                rtl8125_wait_phy_ups_resume(hw, 2);
        //                rtl8125_clear_ups_resume_bit(hw);
        //        }
        //}

        //omit fiber nic setting

        hw->phy_reg_anlpar = 0;
}

void rtl8125_disable_aspm_clkreq_internal(struct rtl_hw *hw)
{
        switch (hw->mcfg) {
        case CFG_METHOD_2 ... CFG_METHOD_5:
                rtl8125_enable_cfg9346_write(hw);
                RTL_W8(hw, Config5, RTL_R8(hw, Config5) & ~BIT_0);
                RTL_W8(hw, Config2, RTL_R8(hw, Config2) & ~BIT_7);
                rtl8125_disable_cfg9346_write(hw);
                RTL_W8(hw, 0xF1, RTL_R8(hw, 0xF1) & ~BIT_7);
        }
}

void rtl8125_disable_ups(struct rtl_hw *hw)
{
        switch (hw->mcfg) {
        case CFG_METHOD_2 ... CFG_METHOD_5:
                rtl8125_mac_ocp_write(hw, 0xD40A, rtl8125_mac_ocp_read( hw, 0xD40A) & ~(BIT_4));
        }
}

void
rtl8125_init_software_variable(struct rtl_hw *hw)
{
        unsigned int speed_mode = SPEED_2500;
        unsigned int duplex_mode = DUPLEX_FULL;
        unsigned int autoneg_mode = AUTONEG_ENABLE;
        unsigned int advertising_mode = ADVERTISE_10_HALF |
                                        ADVERTISE_100_FULL |
                                        ADVERTISE_100_HALF |
                                        ADVERTISE_100_FULL |
                                        ADVERTISE_1000_HALF |
                                        ADVERTISE_1000_FULL |
                                        ADVERTISE_2500_FULL;

        switch (hw->mcfg) {
        case CFG_METHOD_2 ... CFG_METHOD_5:
                hw->HwSuppNowIsOobVer = 1;
        }

        switch (hw->mcfg) {
        case CFG_METHOD_2 ... CFG_METHOD_5:
                hw->HwSuppCheckPhyDisableModeVer = 3;
        }

        switch (hw->mcfg) {
        case CFG_METHOD_2 ... CFG_METHOD_5:
                hw->HwSuppTxNoCloseVer = 3;
        }

        if (hw->HwSuppTxNoCloseVer > 0)
                hw->EnableTxNoClose = TRUE;//TRUE

        switch (hw->mcfg) {
        case CFG_METHOD_2 ... CFG_METHOD_5:
                hw->HwSuppMacMcuVer = 2;
                break;
        }

        switch (hw->mcfg) {
        case CFG_METHOD_2 ... CFG_METHOD_5:
                hw->MacMcuPageSize = RTL8125_MAC_MCU_PAGE_SIZE;
                break;
        }

        switch (hw->mcfg) {
        case CFG_METHOD_2:
                hw->sw_ram_code_ver = NIC_RAMCODE_VERSION_CFG_METHOD_2;
                break;
        case CFG_METHOD_3:
                hw->sw_ram_code_ver = NIC_RAMCODE_VERSION_CFG_METHOD_3;
                break;
        case CFG_METHOD_4:
                hw->sw_ram_code_ver = NIC_RAMCODE_VERSION_CFG_METHOD_4;
                break;
        case CFG_METHOD_5:
                hw->sw_ram_code_ver = NIC_RAMCODE_VERSION_CFG_METHOD_5;
                break;
        }
        rtl8125_link_option((u8*)&autoneg_mode, (u32*)&speed_mode, (u8*)&duplex_mode, (u32*)&advertising_mode);
        hw->autoneg = autoneg_mode;
        hw->speed = speed_mode;
        hw->duplex = duplex_mode;
        hw->advertising = advertising_mode;
}

void
rtl8125_get_tally_stats(struct rtl_hw *hw, struct rte_eth_stats *rte_stats)
{
        struct rtl8125_counters *counters;
        uint64_t paddr;
        u32 cmd;
        u32 WaitCnt;

        counters = hw->tally_vaddr;
        paddr = hw->tally_paddr;
        if (!counters)
                return;

        RTL_W32(hw, CounterAddrHigh, (u64)paddr >> 32);
        cmd = (u64)paddr & DMA_BIT_MASK(32);
        RTL_W32(hw, CounterAddrLow, cmd);
        RTL_W32(hw, CounterAddrLow, cmd | CounterDump);

        WaitCnt = 0;
        while (RTL_R32(hw, CounterAddrLow) & CounterDump) {
                udelay(10);

                WaitCnt++;
                if (WaitCnt > 20)
                        break;
        }

        /* Rx Errors */
        rte_stats->imissed = rte_le_to_cpu_64(counters->rx_missed);
        rte_stats->ierrors = rte_le_to_cpu_64(counters->rx_errors);

        /* Tx Errors */
        rte_stats->oerrors = rte_le_to_cpu_64(counters->tx_errors);

        rte_stats->ipackets = rte_le_to_cpu_64(counters->rx_packets);
        rte_stats->opackets = rte_le_to_cpu_64(counters->tx_packets);

}

void
rtl8125_clear_tally_stats(struct rtl_hw *hw)
{
        if (!hw->tally_paddr)
                return;

        RTL_W32(hw, CounterAddrHigh, (u64)hw->tally_paddr >> 32);
        RTL_W32(hw, CounterAddrLow, ((u64)hw->tally_paddr & (DMA_BIT_MASK(32))) | CounterReset);

}

void
rtl8125_switch_mac_mcu_ram_code_page(struct rtl_hw *hw, u16 page)
{
        u16 tmpUshort;

        page &= (BIT_1 | BIT_0);
        tmpUshort = rtl8125_mac_ocp_read(hw, 0xE446);
        tmpUshort &= ~(BIT_1 | BIT_0);
        tmpUshort |= page;
        rtl8125_mac_ocp_write(hw, 0xE446, tmpUshort);
}

void
_rtl8125_write_mac_mcu_ram_code(struct rtl_hw *hw, const u16 *entry, u16 entry_cnt)
{
        u16 i;

        for (i = 0; i < entry_cnt; i++) {
                rtl8125_mac_ocp_write(hw, 0xF800 + i * 2, entry[i]);
        }
}

void
_rtl8125_write_mac_mcu_ram_code_with_page(struct rtl_hw *hw, const u16 *entry, u16 entry_cnt, u16 page_size)
{
        u16 i;
        u16 offset;

        if (page_size == 0) return;

        for (i = 0; i < entry_cnt; i++) {
                offset = i % page_size;
                if (offset == 0) {
                        u16 page = (i / page_size);
                        rtl8125_switch_mac_mcu_ram_code_page(hw, page);
                }
                rtl8125_mac_ocp_write(hw, 0xF800 + offset * 2, entry[i]);
        }
}

void
rtl8125_write_mac_mcu_ram_code(struct rtl_hw *hw, const u16 *entry, u16 entry_cnt)
{
        if (FALSE == HW_SUPPORT_MAC_MCU(hw)) return;
        if (entry == NULL || entry_cnt == 0) return;

        if (hw->MacMcuPageSize > 0)
                _rtl8125_write_mac_mcu_ram_code_with_page(hw, entry, entry_cnt, hw->MacMcuPageSize);
        else
                _rtl8125_write_mac_mcu_ram_code(hw, entry, entry_cnt);
}
